==> sonusai-0.8.1.tar.gz: sonusai-0.8.1/sonusai/data_generator/keras_from_h5.py <==
from dataclasses import dataclass
from os import PathLike
from typing import Union

import numpy as np
from tensorflow.keras.utils import Sequence

from sonusai.mixture.mixdb import MixtureDatabase
from sonusai.mixture.mixdb import MixtureID


class KerasFromH5(Sequence):
    """Generates data for Keras from a directory containing genft data"""

    @dataclass(frozen=True)
    class BatchParams:
        mixid: MixtureID
        offset: int
        extra: int

    def __init__(self,
                 mixdb: MixtureDatabase,
                 mixid: MixtureID,
                 location: Union[str, bytes, PathLike],
                 batch_size: int,
                 timesteps: int,
                 flatten: bool,
                 add1ch: bool,
                 shuffle: bool = False):
        """Initialization"""
        from os.path import exists
        from os.path import isdir
        from os.path import join

        from sonusai import SonusAIError
        from sonusai.mixture.feature import get_feature_stats
        from sonusai.mixture.mixdb import convert_mixid_to_list
        from sonusai.mixture.mixdb import get_mixid_padded_name

        self.mixdb = mixdb
        self.mixid = convert_mixid_to_list(self.mixdb, mixid)
        self.location = location
        self.batch_size = batch_size
        self.timesteps = timesteps
        self.flatten = flatten
        self.add1ch = add1ch
        self.shuffle = shuffle

        if not isdir(self.location):
            raise SonusAIError(f'Specified location, {self.location}, is not a directory')

        for m in self.mixid:
            mixid_padded_name = get_mixid_padded_name(self.mixdb, m)
            filename = join(self.location, mixid_padded_name + '_feature.h5')
            if not exists(filename):
                raise SonusAIError(f'Could not find {filename}')
            filename = join(self.location, mixid_padded_name + '_truth_f.h5')
            if not exists(filename):
                raise SonusAIError(f'Could not find {filename}')

        fs = get_feature_stats(feature_mode=self.mixdb.feature,
                               frame_size=self.mixdb.frame_size,
                               num_classes=self.mixdb.num_classes,
                               truth_mutex=self.mixdb.truth_mutex)

        self.stride = fs.stride
        self.num_bands = fs.num_bands
        self.num_classes = self.mixdb.num_classes
        self.mixtures = None
        self.mixture_frame_segments = None
        self.batch_frame_segments = None
        self.total_batches = None

        self._initialize_mixtures()

    def __len__(self) -> int:
        """Denotes the number of batches per epoch"""
        return self.total_batches

    def __getitem__(self, batch_index: int) -> (np.ndarray, np.ndarray):
        """Get one batch of data"""
        from sonusai.utils.reshape import reshape_inputs
        from sonusai.utils.parallel import p_map

        batch_params = self.batch_params[batch_index]

        result = p_map(self._get_ft, batch_params.mixid)

        feature = np.vstack([result[i][0] for i in range(len(result))])
        truth = np.vstack([result[i][1] for i in range(len(result))])

        if batch_params.extra > 0:
            feature = feature[batch_params.offset:-batch_params.extra]
            truth = truth[batch_params.offset:-batch_params.extra]
        else:
            feature = feature[batch_params.offset:]
            truth = truth[batch_params.offset:]

        feature, truth, _, _, _, _ = reshape_inputs(feature=feature,
                                                    truth=truth,
                                                    batch_size=self.batch_size,
                                                    timesteps=self.timesteps,
                                                    flatten=self.flatten,
                                                    add1ch=self.add1ch)

        return feature, truth

    def on_epoch_end(self) -> None:
        """Modification of dataset between epochs"""
        import random

        if self.shuffle:
            random.shuffle(self.mixid)
            self.initialize_mixtures()

    def _get_ft(self, mixid: int) -> (np.ndarray, np.ndarray):
        from os.path import join

        import h5py

        from sonusai.mixture.mixdb import get_mixid_padded_name

        mixid_padded_name = get_mixid_padded_name(self.mixdb, mixid)

        filename = join(self.location, mixid_padded_name + '_feature.h5')
        with h5py.File(filename, 'r') as f:
            feature = np.array(f['feature'])

        filename = join(self.location, mixid_padded_name + '_truth_f.h5')
        with h5py.File(filename, 'r') as f:
            truth_f = np.array(f['truth_f'])

        return feature, truth_f

    def _initialize_mixtures(self) -> None:
        from sonusai.data_generator.keras_utils import get_frames_per_batch
        from sonusai.mixture.mixdb import get_mixtures_from_mixid

        self.mixtures = get_mixtures_from_mixid(mixdb=self.mixdb, mixid=self.mixid)
        (frames_per_batch,
         self.total_batches) = get_frames_per_batch(mixdb=self.mixdb,
                                                    mixtures=self.mixtures,
                                                    batch_size=self.batch_size,
                                                    timesteps=self.timesteps)

        # Compute mixid, offset, and extra for dataset
        # offsets and extras are needed because mixtures are not guaranteed to fall on batch boundaries.
        # When fetching a new index that starts in the middle of a sequence of mixtures, the
        # previous feature frame offset must be maintained in order to preserve the correct
        # data sequence. And the extra must be maintained in order to preserve the correct data length.
        cumulative_frames = 0
        start_mixture_index = 0
        offset = 0
        self.batch_params = []
        for m in range(len(self.mixid)):
            current_frames = self.mixtures[m].samples // self.mixdb.feature_step_samples
            cumulative_frames += current_frames
            while cumulative_frames >= frames_per_batch:
                extra = cumulative_frames - frames_per_batch
                self.batch_params.append(self.BatchParams(mixid=self.mixid[start_mixture_index:m + 1],
                                                          offset=offset,
                                                          extra=extra))
                if extra == 0:
                    start_mixture_index = m + 1
                    offset = 0
                else:
                    start_mixture_index = m
                    offset = current_frames - extra
                cumulative_frames = extra

==> sonusai-0.8.2.tar.gz: sonusai-0.8.2/sonusai/data_generator/keras_from_h5.py <==
from dataclasses import dataclass
from os import PathLike
from typing import List
from typing import Union

import numpy as np
from tensorflow.keras.utils import Sequence

from sonusai.mixture.mixdb import MixtureDatabase
from sonusai.mixture.mixdb import MixtureIDs


class KerasFromH5(Sequence):
    """Generates data for Keras from a directory containing genft data"""

    @dataclass(frozen=True)
    class BatchParams:
        mixid: List[int]
        offset: int
        extra: int

    def __init__(self,
                 mixdb: MixtureDatabase,
                 mixids: MixtureIDs,
                 location: Union[str, bytes, PathLike],
                 batch_size: int,
                 timesteps: int,
                 flatten: bool,
                 add1ch: bool,
                 shuffle: bool = False):
        """Initialization"""
        from os.path import exists
        from os.path import isdir
        from os.path import join

        import h5py

        from sonusai import SonusAIError
        from sonusai.mixture.feature import get_feature_stats
        from sonusai.mixture.mixdb import convert_mixids_to_list
        from sonusai.mixture.mixdb import get_mixid_padded_name

        self.mixdb = mixdb
        self.mixids = convert_mixids_to_list(self.mixdb, mixids)
        self.location = location
        self.batch_size = batch_size
        self.timesteps = timesteps
        self.flatten = flatten
        self.add1ch = add1ch
        self.shuffle = shuffle

        if not isdir(self.location):
            raise SonusAIError(f'Specified location, {self.location}, is not a directory')

        for m in self.mixids:
            mixid_padded_name = get_mixid_padded_name(self.mixdb, m)
            filename = join(self.location, mixid_padded_name + '.h5')
            if not exists(filename):
                raise SonusAIError(f'Could not find {filename}')
            with h5py.File(filename, 'r') as f:
                datasets = ['feature', 'truth_f']
                for dataset in datasets:
                    if dataset not in f:
                        raise SonusAIError(f'Could not find "{dataset}" in {filename}')

        fs = get_feature_stats(feature_mode=self.mixdb.feature,
                               frame_size=self.mixdb.frame_size,
                               num_classes=self.mixdb.num_classes,
                               truth_mutex=self.mixdb.truth_mutex)

        self.stride = fs.stride
        self.num_bands = fs.num_bands
        self.num_classes = self.mixdb.num_classes
        self.mixtures = None
        self.mixture_frame_segments = None
        self.batch_frame_segments = None
        self.total_batches = None

        self._initialize_mixtures()

    def __len__(self) -> int:
        """Denotes the number of batches per epoch"""
        return self.total_batches

    def __getitem__(self, batch_index: int) -> (np.ndarray, np.ndarray):
        """Get one batch of data"""
        from sonusai.utils.reshape import reshape_inputs

        batch_params = self.batch_params[batch_index]

        result = []
        for mixid in batch_params.mixid:
            result.append(self._get_ft(mixid))

        feature = np.vstack([result[i][0] for i in range(len(result))])
        truth = np.vstack([result[i][1] for i in range(len(result))])

        if batch_params.extra > 0:
            feature = feature[batch_params.offset:-batch_params.extra]
            truth = truth[batch_params.offset:-batch_params.extra]
        else:
            feature = feature[batch_params.offset:]
            truth = truth[batch_params.offset:]

        feature, truth, _, _, _, _ = reshape_inputs(feature=feature,
                                                    truth=truth,
                                                    batch_size=self.batch_size,
                                                    timesteps=self.timesteps,
                                                    flatten=self.flatten,
                                                    add1ch=self.add1ch)

        return feature, truth

    def on_epoch_end(self) -> None:
        """Modification of dataset between epochs"""
        import random

        if self.shuffle:
            random.shuffle(self.mixids)
            self._initialize_mixtures()

    def _get_ft(self, mixid: int) -> (np.ndarray, np.ndarray):
        from os.path import join

        import h5py

        from sonusai.mixture.mixdb import get_mixid_padded_name

        mixid_padded_name = get_mixid_padded_name(self.mixdb, mixid)

        with h5py.File(join(self.location, mixid_padded_name + '.h5'), 'r') as f:
            feature = np.array(f['feature'])
            truth_f = np.array(f['truth_f'])

        return feature, truth_f

    def _initialize_mixtures(self) -> None:
        from sonusai.data_generator.keras_utils import get_frames_per_batch
        from sonusai.mixture.mixdb import get_mrecords_from_mixids

        self.mixtures = get_mrecords_from_mixids(mixdb=self.mixdb, mixids=self.mixids)
        frames_per_batch, self.total_batches = get_frames_per_batch(mixdb=self.mixdb,
                                                                    mixtures=self.mixtures,
                                                                    batch_size=self.batch_size,
                                                                    timesteps=self.timesteps)

        # Compute mixid, offset, and extra for dataset
        # offsets and extras are needed because mixtures are not guaranteed to fall on batch boundaries.
        # When fetching a new index that starts in the middle of a sequence of mixtures, the
        # previous feature frame offset must be maintained in order to preserve the correct
        # data sequence. And the extra must be maintained in order to preserve the correct data length.
        cumulative_frames = 0
        start_mixture_index = 0
        offset = 0
        self.batch_params = []
        for m in range(len(self.mixids)):
            current_frames = self.mixtures[m].samples // self.mixdb.feature_step_samples
            cumulative_frames += current_frames
            while cumulative_frames >= frames_per_batch:
                extra = cumulative_frames - frames_per_batch
                mixid = self.mixids[start_mixture_index:m + 1]
                self.batch_params.append(self.BatchParams(mixid=mixid, offset=offset, extra=extra))
                if extra == 0:
                    start_mixture_index = m + 1
                    offset = 0
                else:
                    start_mixture_index = m
                    offset = current_frames - extra
                cumulative_frames = extra
